"""Binary data communication model implementations for Kaira.

This module provides models specifically designed for binary data transmission.
"""

from . import soft_bit_thresholding

__all__ = [
    "soft_bit_thresholding",
]
